package projet_these.et;

import java.io.Serializable;
import java.lang.reflect.Array;
import java.util.*;
import projet_these.rn.poidev;




/**
 * Implementation of a population
 * @author Gregory Deceliere - UMR CNRS 5558 - "BBE"
 * @version 01/03/02
 */
public class Population implements Serializable
{
    
    /**
     * Population size
     */
    public int taille;
    
    /**
     * Close populations
     */
    protected Population[] voisins;
    
    /**
     * Migration rate between each close population and the current one
     */
    protected int[] taux_migration;
    
    /**
     * Maximu fitness coefficient initialized to 0
     */
    protected float wmax = 0f;
	
    /**
     * Default alpha selection parameter value
     */
    public final static float DEFAULT_ALPHA = 1E-4f;

    /**
     * Default beta selection parameter value
     */
    public final static float DEFAULT_BETA = 1E-4f;

    /**
     * alpha selection parameter value
     */
    private static float alpha = DEFAULT_ALPHA;

    /**
     * beta selection parameter value
     */
    private static float beta = DEFAULT_BETA;
        
    /**
     * he vector containing the individuals of the population
     */
    protected Vector population;
    
    /**
     * The vector containing the imigrant individuals
     */
    protected Vector migrants;
    
    /**
     * Array of the average population copies number in funtcion of time
     */
    public float[] nb_cop;
    
    /**
     * Array of the variance of population copies number in function of time
     */
    public float[] var_nb_cop;
    
    /**
     * Copies Number profile
     */
    protected Vector profile;
    
    /**
     * The Vector containing the insertion profiles during equilibrium period
     */
    public Vector ins_profile;
    
    
    /**
     * Contructor of initialization of the population
     * @param t Population size
     * @param g Number of simulated generations
     */
    public Population(int t, int g)
    {
        // Initialiser la taille de la population
        taille = t;
        
        /* Initialiser la taille du tableau des nombre moyen de copie*/
        nb_cop = new float[g];
        
        // Initialisation de la population et des imigrants
        population = new Vector();
        migrants = new Vector();
    }
    
    
    /**
     * Contructor of initialization of the population
     * @param t Population size
     * @param n Average number of copies at initial time
     * @param g Number of simulated generations
     */
    public Population(int t, float n, int g)
    {
        this(t, g);
        
        // Initialisation des individus composant la population
        for (int i = 0; i < t; i++)
            population.add(new Individu((int) (poidev.poisson(n))));
        
        // Initialisation du profile des copies
        setFrequencies();
        
        // Initialisation du wmax
        setWmax();
        
        // Initialisation de la veritable moyenne initiale
        moyenne(0);
    }
    
    
    /**
     * Procedure of reproduction of individuals monoecious diploids
     * excluding self fertilization
     */
    protected void reproduction()
    {
        // Individus impliqus dans la reproduction
        Individu[] parents = new Individu[2];
        
        // Initialisation du vecteur de stockage de la population des zygotes
        Vector new_pop;
        new_pop = new Vector();
        
        // Renouvellement de la population par reproduction sexuee
        for (int i = 0; i < taille; i++)
        {
            coupling(parents);
            Individu zygote =
                    fecundation(parents[0], parents[1],
                    parents[0].getClass());
            new_pop.add(zygote);
        }
        
        // Mort des adultes et passage au stade adulte des zygotes
        population = new_pop;
    }
    
    
    /**
     * Procedure of imigration
     */
    protected void migration()
    {
        
        /* Initialisation du vecteur des migrants et mise  jour des populations
        voisines*/
        for (int i = 0; i < taux_migration.length; i++)
        {
            for (int j = 0; j < taux_migration[i]; j++)
            {
                
                /* Tirage uniforme sans remise du migrant dans
                la population voisine i*/
                int ra = (int)(Math.random() *
                        (voisins[i].population.size()));
                
                /* Remarque 1 : le tirage uniforme d'un nombre x element de
                                                [alpha,beta] se fait par :
                                                alpha+(beta-alpha+1)*Math.random()*/
                
                /* Ajout de l'individu dans la population courante
                mais suppression chez la voisine
                 */
                migrants.add(
                        (Individu) voisins[i].population.elementAt(ra));
                voisins[i].population.remove(ra);
            }
        }
    }
    
    
    /**
     * Fusion procedure of the vector of the individuals imigrants and
     * that of the remainders leading to the formation of
     * the new current population
     * @param temps The current generation
     */
    protected void post_migration(int temps)
    {
        // Ajout des migrants aux restants
        population.addAll(migrants);
        
        // Liberation des individus comptenus dans le vecteur des migrants
        migrants.removeAllElements();
        
        // Renouvellement du profile des copies
        setFrequencies();
        
        // Initialisation du wmax
        setWmax();
        
        // Calcul de la nouvelle moyenne
        moyenne(temps);
    }
    
    
    /**
     * Procedure of coupling between two individus
     * @param parents     The parents
     */
    protected void coupling(Individu[] parents)
    {
        // Numeros associes aux deux individus impliques dans la reproduction
        int ra1, ra2;
        
        /* Tirage uniforme du premier parent dans la population des adultes
        et selection*/
        do
        {
            ra1 = (int)(Math.random() * taille);
            parents[0] = (Individu) population.elementAt(ra1);
        }
        while (Math.random() > (fitness(parents[0]) / wmax))
            ;
        
        /* Tirage uniforme du second parent dans la population des adultes
        privee du premier parent et selection*/
        do
        {
            ra2 = (int)(Math.random() * taille);
            parents[1] = (Individu) population.elementAt(ra2);
        }
        while ((ra2 == ra1) ||
                (Math.random() > (fitness(parents[1]) / wmax)));
    }
    
    
    /**
     * Function comptuting the numerator of a genetical type fitness
     * @param ind     Individual characterized by a genetical type
     * @return        The fitness coefficient wi
     */
    protected float fitness(Individu ind)
    {

        // Selection parameters
        //float alpha = 1E-4f, beta = 1E-4f;

        // Genetical type i
        int ni = ind.getEntite_sel();

        // Return fitness of the individu
        return (float) Math.exp((double) - ((float) ni) *
                                (Population.alpha + Population.beta * ((float) ni) / 2f));
    }
    
    
    /**
     * Procedure of fecundation between two gamete generate by
     * each parent 
     * @return The Zygote resulting from the fecundation
     * @param ind_class Class
     * @param ind1 First parent
     * @param ind2 Other parent
     */
    private Individu fecundation(Individu ind1, Individu ind2, Class ind_class)
    {
        // Paires de chromosomes du zygote
        ChromosomesPair[] chromo_pairs = new ChromosomesPair[3];
        
        // Classe des paires de chromosomes
        Class chromo_pairs_class;
        
        // Gamete du premier individu implique dans la formation du zygote
        Hashtable[] gamete_1;
        
        // Gamete du second individu implique dans la formation du zygote
        Hashtable[] gamete_2;
        
        // Zygote
        Individu zygote = new Individu();
        
        // Formation des gametes parentaux
        gamete_1 = ind1.gametes();
        gamete_2 = ind2.gametes();
        
        // Formation du zygote
        try
        {
            chromo_pairs_class = ind_class.
                    getDeclaredField("chromo_pairs_class").getType();
            chromo_pairs = (ChromosomesPair[])
            Array.newInstance(chromo_pairs_class, 3);
            zygote = (Individu) ind_class.newInstance();
            for (int i = 0; i < 3; i++)
            {
                chromo_pairs[i] =
                        (ChromosomesPair) chromo_pairs_class.newInstance();
                chromo_pairs[i] =
                        chromo_pairs[i].setChromo(gamete_1[i], gamete_2[i]);
            }
        }
        catch(Exception e)
        {};
        
        return zygote.setChromoPairs(chromo_pairs[0], chromo_pairs[1],
                chromo_pairs[2]);
    }
    
    
    /**
     * Calculation procedure of the average number of copies to a generation
     * temps
     * @param temps The current generation
     */
    protected void moyenne(int temps)
    {
        
        // Initialisation du nombre moyen de copies
        float mean = 0f;
        
        // Calcul de la somme
        for (int i = 0; i < profile.size(); i++)
            mean += ((float) i) * ((Integer) profile.get(i)).floatValue();
        
        // Calcul de la moyenne
        nb_cop[temps] = mean / (float) taille;
    }
    
    
    /**
     * Calculation procedure of the variance of copies number to a generation
     * temps
     * @param temps The current generation
     */
    public void variance(int temps)
    {
        // Initialisation de la moyenne du crre du nombre de copies
        float mean2 = 0f;
        
        for (int i = 0; i < profile.size(); i++)
            mean2 += ((float) (i * i)) * ((Integer) profile.get(i)).
                    floatValue();
        mean2 /= (float) taille;
        mean2 -= nb_cop[temps] * nb_cop[temps];
        
        /* Initialise si necessaire la taille du tableau des
        variances du nombre de copie*/
        if (this.var_nb_cop == null)
            this.var_nb_cop = new float[this.nb_cop.length - temps];
        
        var_nb_cop[-this.nb_cop.length + this.var_nb_cop.length +
                temps] = mean2;
    }
    
    
    /**
     * Procedure setting the profile of insertion sites
     * @param temps The current generation
     */
    public void setIns_Profile(int temps)
    {
        // Current Individu drawn from population
        Individu ind;
        
        // ChromosomesPair of the current Individu
        ChromosomesPair chromo;
        
        // Position along ChromosomesPair
        Byte pos;
        
        // Insertion profile at generation temps
        int[][] table = new int[3][120];
        
        for (int i = 0; i < 3; i++)
        {
            Arrays.fill(table[i], 0);
            for (int j = 0; j < population.size(); j++)
            {
                ind = (Individu) population.elementAt(j);
                chromo = ind.XX[i];
                for (int k = 0; k < 120; k++)
                {
                    pos = new Byte((byte) (k + 1));
                    table[i][k] += (chromo.inser_a.containsKey(pos)) ?
                        1 : 0;
                    table[i][k] += (chromo.inser_b.containsKey(pos)) ?
                        1 : 0;
                }
            }
        }
        /* Initialise si necessaire le vecteur des profiles d'insertion*/
        if (this.ins_profile == null)
            this.ins_profile = new Vector(this.nb_cop.length - temps);
        
        // Storage of this insertion profile
        ins_profile.add(table);
    }
    
    
    /**
     * Procedure setting the profile of the population
     */
    protected void setFrequencies()
    {
        Vector freq = new Vector();
        setFrequencies(this.population, freq);
        this.profile = freq;
    }
    
    
    /**
     * Procedure setting a profile
     * @param pop       The population on witch the profile is made
     * @param freq      The profile of copy number in the population
     */
    protected void setFrequencies(Vector pop, Vector freq)
    {
        // Index de classe dans le vecteur profile
        int index;
        
        // Initialisation du vecteur avec des zeros
        freq.setSize(100);
        for (int i = 0; i < freq.size(); i++)
            freq.setElementAt(new Integer(0), i);
        
        // Construction du vecteur profile
        for (int i = 0; i < pop.size(); i++)
        {
            index = ((Individu) pop.elementAt(i)).nb_copies;
            if(index >= freq.size())
            {
                int j = freq.size();
                freq.setSize(index + 1);
                for (; j < freq.size(); j++)
                    freq.setElementAt(new Integer(0), j);
            }
            freq.setElementAt(new Integer(((Integer) freq.
                    elementAt(index)).intValue() + 1), index);
        }
    }
    
    
    /**
     * Procedure setting the population's wmax
     */
    protected void setWmax()
    {
        wmax = 0;
        for (int i = 0; i < taille; i++)
            wmax = (fitness((Individu) population.elementAt(i))
            > wmax) ? fitness((Individu) population.elementAt(i))
            : wmax;
    }
    
    
    /**
     * Initialization procedure of the migration rates
     * @param taux_migration The table of migration rates
     */
    public void setTaux_Migration(int[] taux_migration)
    {
        
        this.taux_migration = taux_migration;
    }
    
    
    /**
     * Initialization procedure of the neighbors
     * @param voisins The neighbors
     */
    public void setVoisins(Population[] voisins)
    {
        
        this.voisins = voisins;
    }
    
    
    /**
     * Procedure setting the alpha value
     * @param alpha value
     */
    public static void setAlphaValue(float alpha_value)
    {

        Population.alpha = alpha_value;
    }

    /**
     * Procedure setting the alpha value
     * @param alpha value
     */
    public static void setBetaValue(float beta_value)
    {

        Population.beta = beta_value;
    }

    
    /**
     * Functin returning all copies number
     * @param   generation  The period in witch copies number would be return
     * @return              Total number of copies
     */
    public float getAllCopiesNb(int generation)
    {
        return this.nb_cop[generation];
    }    

    
    /**
     * Functin returning alpha value
     * @return              alpha value
     */
    static public float getAlphaValue()
    {
        return Population.alpha;
    }

    /**
     * Functin returning beta value
     * @return              beta value
     */
    static public float getBetaValue()
    {
        return Population.beta;
    }
    
}


